import os
import re
import logging

from typing import Optional

log = logging.getLogger(__name__)


def safe_filename(title: str) -> str:
    pattern = re.compile(r'[?"|:<>*/\\]', flags=re.VERBOSE)
    return pattern.sub("", str(title)).strip()


def create_path(path: str) -> str:
    if path:
        if not os.path.isabs(path):
            path = os.path.join(os.getcwd(), path)

    if os.path.exists(path):
        log.warning(f"Path already exists: {path}")
    else:
        log.debug(f"Creating path: {path}")

    os.makedirs(path, exist_ok=True)
    return path


def create_tags(text: str) -> str:
    identifiers = [
        r'is_read_for_free',
        r"is_wait_until_free",
        r"is_already_read",
        r"is_free",
        r"is_zero_plus",
        r"is_purchased",
    ]
    identifiers = "|".join(identifiers)
    pattern = r"(\b" + identifiers + r")\b(\s*(" + identifiers + r")\b)*"

    regex = re.compile(pattern, re.I)
    tags = regex.sub(r"episode['\1']", text.strip('"'))
    return tags


def trunc_title(title: str) -> str:
    return re.sub(r"\((?:[^)(]|\([^)(]*\))*\)", "", title)


def pad_string(
    text: str,
    length: Optional[int] = 0,
    padding: Optional[str] = "0"
) -> str:
    if len(text) <= length:
        for pad in range(length-len(text)):
            padding += "0"
        text = f"{padding}{text}"
    return text



#!/usr/bin/env python

import json
import logging
import requests
from urllib.parse import parse_qs
from typing import Mapping, Union, Dict

from pyccoma import Scraper
from pyccoma.exceptions import PyccomaError, PageError, LoginError
from pyccoma.helpers import trunc_title

from pyccoma.fr.urls import (
    base_url,
    login_url,
    api_url,
    history_url,
    bookmark_url,
    purchase_url,
)

log = logging.getLogger(__name__)


class Pyccoma(Scraper):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.api_url = self.get_api_url()
        self.history_url = history_url % self.api_url
        self.bookmark_url = bookmark_url % self.api_url
        self.purchase_url = purchase_url % self.api_url

        self._etype = {
            "manga": "volume",
            "smartoon": "episode",
            "novel": "episode"
        }

    @property
    def manga(self) -> str:
        return self._etype['manga']

    @property
    def smartoon(self) -> str:
        return self._etype['smartoon']

    @property
    def novel(self) -> str:
        return self._etype['novel']
    
    @manga.setter
    def manga(self, value: str) -> None:
        if value in ('volume', 'episode'):
            self._etype['manga'] = value
        else:
            raise ValueError("Invalid type.")

    @smartoon.setter
    def smartoon(self, value: str) -> None:
        if value in ('volume', 'episode'):
            self._etype['smartoon'] = value
        else:
            raise ValueError("Invalid type.")

    @novel.setter
    def novel(self, value: str) -> None:
        if value in ('volume', 'episode'):
            self._etype['novel'] = value
        else:
            raise ValueError("Invalid type.")

    def parse_json(self, url: str) -> json:
        try:
            page = self.session.get(url, headers=self.headers)
            page.raise_for_status()
            return page.json()

        except requests.exceptions.HTTPError:
            raise PageError(url)
        except requests.exceptions.ConnectionError:
            raise PageError(url)
        except Exception:
            log.error("Failed to parse page.")

    def get_api_url(self) -> str:
        page = self.parse_page(base_url).xpath('//script[@id = "__NEXT_DATA__"]/text()')[0]  # noqa:E501
        build_id = json.loads(page)['buildId']
        new_api_url = api_url % build_id
        return new_api_url

    def login(self, email: str, password: str) -> None:
        try:
            session = self.session.get(login_url, headers=self.headers)

            params = {
                'email': email,
                'password': password,
                'redirect': '/fr/',
            }

            login = self.session.post(
                login_url,
                data=params,
                cookies=session.cookies,
                headers=self.headers
            )

            if login.ok and not 'error' in login.text:
                self._is_login = True
                log.info(f"Successfully logged in as {email}")
            else:
                self._is_login = False
                log.error(f"Failed to log in as {email}")

        except Exception:
            raise SystemExit("Failed to establish connection to server.")

    def get_list(self, url: str) -> Mapping[int, Dict[str, bool]]:
        try:
            return self.get_episode_list(url)
        except Exception:
            raise PageError(url)      

    def get_episode_list(
        self,
        url: str
    ) -> Mapping[int, Dict[str, Union[str, bool]]]:
        try:
            if 'episode' in url:
                type = 'episode'
            elif 'volume' in url:
                type = 'volume'
            else:
                raise ValueError("Invalid url.")

            product_id = url.split("/")[-1]
            url = self.api_url + f"/product/{type}/{product_id}.json?id={product_id}&pathname=%2Fproduct%2F%5Bid%5D"  # noqa:E501
            page = self.parse_json(url)['pageProps']['initialState']['episode']['episodeList']['episode_list']  # noqa:E501
                
            episodes = {
                id + 1: {
                    'title': episode['title'],
                    'url': f"{base_url}/viewer/{product_id}/{episode['id']}",  # noqa:E501
                    'is_free': True if 'FR01' in episode['use_type'] else False,  # noqa:E501
                    'is_read_for_free': True if 'RD01' in episode['use_type'] else False,  # noqa:E501
                    'is_wait_for_free': True if 'WF15' in episode['use_type'] else False,  # noqa:E501
                    'is_purchased': True if 'AB01' in episode['use_type'] else False,  # noqa:E501
                    'is_already_read': episode['is_read'] if self._is_login else False,  # noqa:E501
                }
                for id, episode in enumerate(page)
            }

            if not episodes:
                log.debug(f"No episodes found on {url}.")

            return episodes

        except KeyError:
            log.error("Unable to fetch list.")
        except AttributeError:
            raise PageError(url)
        except Exception as err:
            raise PyccomaError(err)

    def get_bdata(self, url: str) -> Dict[str, str]:
        try:
            page = self.parse_json(url)['pageProps']['initialState']

            if 'history' in url:
                page = page['history']['products']
            elif 'bookmark' in url:
                page = page['bookmark']['products']
            elif 'purchase' in url:
                page = page['purchase']['purchase']['products']

            product_type = [
                self.smartoon if episode['category_id'] == 2 else
                self.novel if episode['category_id'] == 3 else self.manga
                for episode in page
            ]

            items = {
                episode['title']: f"{base_url}/product/{etype}/{episode['id']}"
                for episode, etype in zip(page, product_type)
            }
            
            return items
        
        except Exception:
            log.error("Failed to parse library.")

    def get_pdata(self, url: str) -> Dict[str, Union[str, bool]]:
        try:
            log.info(f"Parsing data from {url}")
            product_id = url.split("/")[-2]
            episode_id = url.split("/")[-1]
            url = self.api_url + f"/viewer/{product_id}/{episode_id}.json?productId={product_id}&episodeId={episode_id}"  # noqa:E501
            page = self.parse_json(url)['pageProps']['initialState']

            authors = ' '.join([author['name'] for author in page['productDetail']['productDetail']['product']['authors']])  # noqa:E501
            title = page['productDetail']['productDetail']['product']['title'] + f" ({authors})"  # noqa:E501
            images = [episode['path'] for episode in page['viewer']['pData']['img']]  # noqa:E501

            pdata = {
                'title': title if not self.omit_author else trunc_title(title),
                'ep_title': page['viewer']['pData']['title'],
                'img': images
            }

            return pdata

        except TypeError or IndexError:
            log.error("Unable to fetch page data.")

    def get_checksum(self, img_url: str) -> str:
        return ' '.join(parse_qs(img_url)['q'])

    def get_history(self) -> Dict[str, str]:
        if self._is_login:
            return self.get_bdata(self.history_url)
        else:
            raise LoginError

    def get_bookmark(self) -> Dict[str, str]:
        if self._is_login:
            return self.get_bdata(self.bookmark_url)
        else:
            raise LoginError

    def get_purchase(self) -> Dict[str, str]:
        if self._is_login:
            return self.get_bdata(self.purchase_url)
        else:
            raise LoginError


#!/usr/bin/env python

import argparse

import os
import re
import logging
from getpass import getpass
from typing import Optional, Tuple, List
from itertools import chain

from pyccoma.jp.pyccoma import Pyccoma as Jp
from pyccoma.fr.pyccoma import Pyccoma as Fr
from pyccoma.exceptions import PyccomaError
from pyccoma.logger import setup_logging, levels
from pyccoma.helpers import create_tags

log = logging.getLogger(__name__)


def main() -> None:
    try:
        global pyccoma, region

        setup_logging()
        parser = construct_parser()
        args = parser.parse_args()

        region = args.region.lower()
        
        if region == 'jp':
            pyccoma = Jp()
        elif region == 'fr':
            pyccoma = Fr()
        else:
            raise ValueError("Invalid country specified.")

        url = args.url
        pyccoma.format = args.format
        pyccoma.manga = args.etype[0]
        pyccoma.smartoon = args.etype[1]
        pyccoma.novel = args.etype[2]
        pyccoma.zeropad = args.pad
        pyccoma.retry_count = args.retry_count
        pyccoma.retry_interval = args.retry_interval
        pyccoma.archive = args.archive
        pyccoma.omit_author = args.omit_author

        logging.getLogger().setLevel(args.loglevel)

        if not (
            args.range and 'viewer' not in args.url
        ) and (
            args.filter == 'custom'
        ):
            raise PyccomaError("Use --filter custom along with --range.")

        elif (
            args.range and not args.filter
        ) or (
            args.range and args.filter in ('min', 'max', 'all')
        ):
            log.warning(f"Overriding --filter={args.filter} to parse custom.")
            args.filter = 'custom'

        password = ""

        if args.email:
            if not password:
                password = getpass()

            pyccoma.login(args.email, password)

        if args.url and args.filter:
            if args.url[0] in ('history', 'bookmark', 'purchase'):
                if args.url[0] in 'history':
                    url = pyccoma.get_history().values()
                    log.info(
                        f"Parsing ({len(url)}) titles from your history."
                    )

                elif args.url[0] in 'bookmark':
                    url = pyccoma.get_bookmark().values()
                    log.info(
                        f"Parsing ({len(url)}) titles from your bookmarks."
                    )

                elif args.url[0] in 'purchase':
                    url = pyccoma.get_purchase().values()
                    log.info(
                        f"Parsing ({len(url)}) titles from your purchases."
                    )

            elif valid_url(args.url[0], level=3):
                raise PyccomaError(
                    "There is nothing to aggregate. You should only use "
                    "--filter on a product page or your library."
                )

        if any(map(valid_url, args.url)):
            if not os.path.exists(args.output) and not args.output:
                log.warning(
                    "No path found, creating an extract folder inside "
                    "the current working directory: {0}".format(os.getcwd())
                )
            fetch(
                url,
                args.filter,
                args.range,
                args.include,
                args.exclude,
                args.output
            )
        else:
            raise ValueError("Invalid url.")

    except Exception as error:
        parser.error(error)


def construct_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="pyccoma",
        description="Scrape and download from Piccoma.",
        epilog="""
        Report bugs and make feature requests at
        https://github.com/catsital/pyccoma/issues
        """,
        add_help=False,
    )

    required = parser.add_argument_group("Required options")
    required.add_argument(
        "url",
        nargs="*",
        help="""
        Link to an episode or product. If logged in, use: history, bookmark,
        or purchase as shorthand to your library.
        """
    )

    optional = parser.add_argument_group("General options")
    optional.add_argument(
        "-o",
        "--output",
        type=str,
        default="extract",
        help="Path to save downloaded images."
    )
    optional.add_argument(
        "-f",
        "--format",
        type=str,
        default="png",
        help="Image format: png, jpeg, jpg, gif, bmp (Default: png)"
    )
    optional.add_argument(
        "-p",
        "--pad",
        type=int,
        metavar=("LENGTH"),
        default=0,
        help="Pad page numbers with leading zeroes (Default: 0)"
    )
    optional.add_argument(
        "--archive",
        dest="archive",
        action="store_true",
        default=False,
        help="Output to cbz archive format."
    )
    optional.add_argument(
        "--omit-author",
        dest="omit_author",
        action="store_true",
        default=False,
        help="Omit author(s) in title naming scheme."
    )

    locale = parser.add_argument_group("Locale options")
    locale.add_argument(
        "--region",
        type=str,
        default="Jp",
        help="Select which Piccoma service to use. (Default: Jp)"
    )

    retry = parser.add_argument_group("Retry options")
    retry.add_argument(
        "--retry-count",
        type=int,
        metavar=("COUNT"),
        default=3,
        help="Number of download retry attempts. (Default: 3)"
    )
    retry.add_argument(
        "--retry-interval",
        type=int,
        metavar=("SECONDS"),
        default=1,
        help="Delay between each retry attempt. (Default: 1)"
    )

    user = parser.add_argument_group("Login options")
    user.add_argument(
        "--email",
        type=str,
        help="Account email address."
    )

    filter = parser.add_argument_group("Filter options")
    filter.add_argument(
        "--etype",
        type=str,
        nargs=3,
        metavar=("MANGA", "SMARTOON", "NOVEL"),
        default=["volume", "episode", "episode"],
        help="""
        Preferred episode type to scrape manga, smartoon and novel when
        aggregating your library. (Default: manga=volume
        smartoon=episode novel=episode)
        """
    )
    filter.add_argument(
        "--filter",
        type=str,
        help="""
        Filter to use when aggregating products or scraping library. Can be
        either: min, max, all, or custom with --range.
        """
    )
    filter.add_argument(
        "--range",
        type=int,
        nargs=2,
        metavar=("START", "END"),
        help="Range to use when scraping episodes or volumes."
    )
    filter.add_argument(
        "--include",
        type=include,
        default="episode",
        help="""
        Arguments to include when parsing your library: is_purchased, is_free,
        is_zero_plus, is_already_read, is_read_for_free, is_wait_until_free
        """
    )
    filter.add_argument(
        "--exclude",
        type=exclude,
        default="",
        help="""
        Arguments to exclude when parsing your library: is_purchased, is_free,
        is_zero_plus, is_already_read, is_read_for_free, is_wait_until_free
        """
    )

    logger = parser.add_argument_group("Logging options")
    logger.add_argument(
        "-l", "--loglevel",
        metavar="LEVEL",
        choices=levels,
        default="info",
        help="""
        Set the log message threshold. Valid levels are: debug, info, warning,
        error, none (Default: info)
        """
    )

    info = parser.add_argument_group("Info")
    info.add_argument(
        "-h", "--help",
        action="help",
        help="Show this help message and exit."
    ),
    info.add_argument(
        "-v", "--version",
        action="version",
        help="Show program version.",
        version="%(prog)s 0.5.1"
    )

    return parser


def include(value: str) -> str:
    try:
        include = create_tags(value).replace("&", " and ").replace("|", " or ")
        return include
    except Exception:
        raise argparse.ArgumentTypeError(
            'Specify type in this format: "is_free|is_already_read"'
        )


def exclude(value: str) -> str:
    try:
        exclude = create_tags(value).replace("&", " and ").replace("|", " or ")
        return exclude
    except Exception:
        raise argparse.ArgumentTypeError(
            'Specify type in this format: "is_free&is_already_read"'
        )


def valid_url(url: str, level: Optional[int] = None) -> bool:
    if region == 'jp':
        base_url = r"(http|https)://(|www.)(|jp.)piccoma.com/web"
        urls = [
            base_url + r"/product/([0-9\-]+)/episodes\?etype\=([eE|vV]+)",
            base_url + r"/product/([0-9\-]+)/episodes\?etype\=([eE]+)",
            base_url + r"/product/([0-9\-]+)/episodes\?etype\=([vV]+)",
            base_url + r"/viewer/(|s/)([0-9]+)/([0-9]+)",
            base_url + r"/bookshelf/|(bookmark|history|purchase)"
        ]
    elif region == 'fr':
        base_url = r"(http|https)://(|www.)(|fr.)piccoma.com/fr"
        urls = [
            base_url + r"/product/([episode|volume]+)/([0-9\-]+)",
            base_url + r"/product/episode/([0-9\-]+)",
            base_url + r"/product/volume/([0-9\-]+)",
            base_url + r"/viewer/([0-9]+)/([0-9]+)",
            base_url + r"/bookshelf/|(bookmark|history|purchase)"
        ]
    else:
        raise ValueError("Invalid input.")

    if not level:
        urls = "|".join(urls)
        regex = re.search(urls, url)
    else:
        regex = re.search(urls[level], url)
    return bool(regex)


def fetch(
    url: List[str],
    mode: Optional[str] = None,
    range: Optional[Tuple[int, int]] = None,
    include: Optional[str] = None,
    exclude: Optional[str] = None,
    output: Optional[str] = None
) -> None:
    if mode:
        try:
            if not range:
                range = (0, 0)

            if exclude:
                exclude = f" {'and ' if include else ''}not ({exclude})"

            product = []

            for title in url:
                product.append([
                    episode['url']
                    for episode in pyccoma.get_list(title).values()
                    if eval((include) + (exclude))
                ])

            if 'min' in mode:
                product = [episode[0] for episode in product if episode]
            elif 'max' in mode:
                product = [episode[-1] for episode in product if episode]
            elif 'all' in mode:
                product = list(chain.from_iterable(product))
            elif 'custom' in mode:
                product = list(chain.from_iterable(product))[range[0]:range[1]]
            else:
                raise ValueError

            log.info("Fetching ({0}) items.".format(total := len(product)))

            for index, item in enumerate(product):
                log.info(f"Fetching ({index+1}/{total})")
                pyccoma.fetch(item, output)

        except Exception as error:
            raise PyccomaError(error)

    else:
        for index, link in enumerate(url):
            if valid_url(url=link, level=3):
                log.info(f"Fetching ({index+1}/{len(url)})")
                pyccoma.fetch(link, output)
            elif valid_url(url=link, level=0):
                raise PyccomaError(
                    "Use --filter to aggregate episodes in product pages."
                )
            else:
                raise ValueError("Invalid url.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import os
import sys
import logging
import zipfile
import requests

from lxml import html
from urllib.parse import parse_qs
from abc import ABCMeta, abstractmethod

from zipfile import ZipFile
from threading import Thread, Lock
from time import time, gmtime, strftime
from requests import get, session, Response
from typing import Optional, Mapping, Union, Dict, List
from functools import lru_cache

from pycasso import Canvas

from pyccoma.exceptions import PyccomaError, PageError
from pyccoma.helpers import create_path, pad_string, safe_filename
from pyccoma.utils import display_progress_bar, retry

log = logging.getLogger(__name__)


class Scraper(metaclass=ABCMeta):
    def __init__(self):
        self.headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) '
                          'AppleWebKit/537.36 (KHTML, like Gecko) '
                          'Chrome/92.0.4515.107 Safari/537.36',
            'Referer': 'https://piccoma.com/'
        }
        self.session = session()
        self.session.verify = True
        self.__is_login = False
        self._lock = Lock()
        self._format = "png"
        self._archive = False
        self._omit_author = False
        self._retry_count = 3
        self._retry_interval = 1
        self._zeropad = 0

    @property
    def format(self) -> str:
        return self._format

    @property
    def archive(self) -> bool:
        return self._archive

    @property
    def omit_author(self) -> bool:
        return self._omit_author

    @property
    def retry_count(self) -> int:
        return self._retry_count

    @property
    def retry_interval(self) -> int:
        return self._retry_interval

    @property
    def zeropad(self) -> int:
        return self._zeropad

    @format.setter
    def format(self, value: str) -> None:
        if value in ('png', 'jpg', 'gif', 'bmp', 'jpeg'):
            self._format = value.lower()
        else:
            raise ValueError("Invalid format.")

    @archive.setter
    def archive(self, value: bool) -> None:
        self._archive = value

    @omit_author.setter
    def omit_author(self, value: bool) -> None:
        self._omit_author = value

    @retry_count.setter
    def retry_count(self, value: int) -> None:
        self._retry_count = value

    @retry_interval.setter
    def retry_interval(self, value: int) -> None:
        self._retry_interval = value

    @zeropad.setter
    def zeropad(self, value: int) -> None:
        self._zeropad = value

    @property
    def _is_login(self) -> bool:
        return self.__is_login

    @_is_login.setter
    def _is_login(self, value: bool) -> None:
        self.__is_login = value

    def parse(self, page: str) -> html:
        return html.fromstring(page)

    def parse_page(self, url: str) -> html:
        try:
            page = self.session.get(url, headers=self.headers)
            page.raise_for_status()
            soup = self.parse(page.text)
            return soup

        except requests.exceptions.HTTPError:
            raise PageError(url)
        except requests.exceptions.ConnectionError:
            raise PageError(url)
        except Exception:
            log.error("Failed to parse page.")
    
    @abstractmethod
    def login(self, email: str, password: str) -> None:
        pass

    @abstractmethod
    def get_list(self, url: str) -> Mapping[int, Dict[str, bool]]:
        pass

    @abstractmethod
    def get_episode_list(
        self,
        url: str
    ) -> Mapping[int, Dict[str, Union[str, bool]]]:
        pass

    @abstractmethod
    def get_bdata(self, url: str) -> Dict[str, str]:
        pass

    @abstractmethod
    def get_pdata(self, url: str) -> Dict[str, Union[str, bool]]:
        pass

    @retry()
    def get_img(self, img_url: str) -> Response:
        try:
            img = get(img_url, headers=self.headers, stream=True)
            return img
        except requests.exceptions.ConnectionError:
            raise Exception(img_url)

    def download(self, img_url: str, seed: str, output: str) -> None:
        try:
            img = self.get_img(img_url)

            if seed.isupper():
                Canvas(img.raw, (50, 50), seed).export(
                    mode="scramble",
                    path=output,
                    format=self.format
                )
            else:
                with open(f"{output}.{self.format}", 'wb') as handler:
                    for chunk in img.iter_content(1024):
                        if chunk:
                            handler.write(chunk)

        except Exception as err:
            log.error(f"Unable to download image. {err}")
        except KeyboardInterrupt:
            pass

    def compress(
        self,
        img_url: str,
        seed: str,
        page: str,
        file: ZipFile
    ) -> None:
        try:
            img = self.get_img(img_url)

            if seed.isupper():
                img = Canvas(img.raw, (50, 50), seed).export(
                    mode="scramble",
                    format=self.format
                )
                img = img.getvalue()
            else:
                img = img.content

            with self._lock:
                file.writestr(page, img)

        except Exception as err:
            log.error(f"Unable to download image. {err}")
        except KeyboardInterrupt:
            pass
        
    def fetch(self, url: str, path: Optional[str] = None) -> None:
        try:
            pdata = self.get_pdata(url)
            sys.stdout.write(
                f"\nTitle: {pdata['title']}\n"
                f"Episode: {pdata['ep_title']}\n"
            )

            if not path:
                path = os.path.join(os.getcwd(), 'extract')

            start_time = time()
            self._fetch(pdata['img'], pdata['title'], pdata['ep_title'], path)

            with self._lock:
                exec_time = strftime("%H:%M:%S", gmtime(time() - start_time))
                sys.stdout.write(f"\nElapsed time: {exec_time}\n\n")
                sys.stdout.flush()

        except TypeError:
            log.error("Unable to fetch episode.")
        except IndexError:
            log.error(f"Unable to access page on {url}")
        except Exception as err:
            raise PyccomaError(err)
        except KeyboardInterrupt:
            pass

    def _fetch(
        self,
        episode: List[str],
        title: str,
        ep_title: str,
        path: str
    ) -> None:
        try:
            count = 0
            episode_size = len(episode)
            checksum = self.get_checksum(episode[0])
            key = self.get_key(episode[0])
            seed = self.get_seed(checksum, key)
            title = safe_filename(title)
            ep_title = safe_filename(ep_title)

            if not self.archive:
                head_path = os.path.join(path, f"{title}/{ep_title}/")
                path = create_path(head_path)
            else:
                head_path = os.path.join(path, f"{title}_{ep_title}.cbz")
                path = create_path(path)

                if os.path.exists(head_path):
                    log.warning(f"File already exists: {head_path}")

                file = ZipFile(head_path, "a", zipfile.ZIP_DEFLATED, False)

            for page, url in enumerate(episode):
                output = os.path.join(
                    path,
                    page := pad_string(str(page + 1), length=self.zeropad)
                )

                if not self.archive and os.path.exists(file_name := f"{output}.{self.format}"):  # noqa:E501
                    log.debug(f"File already exists: {file_name}")
                elif self.archive and (file_name := f"{page}.{self.format}") in file.namelist():  # noqa:E501
                    log.debug(f"File already exists: {file_name}")
                else:
                    if self.archive:
                        fetch = Thread(
                            target=self.compress,
                            args=(url, seed, file_name, file)
                        )
                    else:
                        fetch = Thread(
                            target=self.download,
                            args=(url, seed, output)
                        )
                    fetch.start()

                with self._lock:
                    count += 1
                    display_progress_bar(count, episode_size)

        except Exception as err:
            log.error(f"Unable to fetch episode. {err}")
        except KeyboardInterrupt:
            pass
    
    @abstractmethod
    def get_checksum(img_url: str) -> str:
        pass

    def get_key(self, img_url: str) -> str:
        return ' '.join(parse_qs(img_url)['expires'])

    def get_seed(self, checksum: str, expiry_key: int) -> str:
        for num in expiry_key:
            if int(num) != 0:
                checksum = checksum[-int(num):] + checksum[:len(checksum)-int(num)]
        return checksum

    @abstractmethod
    @lru_cache
    def get_history(self) -> Dict[str, str]:
        pass

    @abstractmethod
    @lru_cache
    def get_bookmark(self) -> Dict[str, str]:
        pass

    @abstractmethod
    @lru_cache
    def get_purchase(self) -> Dict[str, str]:
        pass



import sys
import logging

from time import sleep
from typing import Callable
from functools import wraps
from requests import Response

log = logging.getLogger(__name__)


def retry() -> Callable[..., Response]:
    def _retry(func):
        @wraps(func)
        def download(self, *args, **kwargs):
            with self._lock:
                try:
                    for retry in range(1, self.retry_count + 1):
                        try:
                            return func(self, *args, **kwargs)
                        except Exception as err:
                            if retry == self.retry_count:
                                raise Exception from err
                            else:
                                log.error(
                                    f"Retrying ({retry}/{self.retry_count}) "
                                    f"{err}"
                                )
                                sleep(self.retry_interval)
                except Exception:
                    log.error(
                        f"Maximum retries exceeded ({retry}/"
                        f"{self.retry_count})"
                    )
        return download
    return _retry


def display_progress_bar(
    file_count: int,
    total_count: int,
    char: str = "█",
    scale: float = 0.55
) -> None:
    max_width = int(100 * scale)
    filled = int(round(max_width * file_count / float(total_count)))
    remaining = max_width - filled
    progress_bar = char * filled + " " * remaining
    percent = round(100.0 * file_count / float(total_count), 1)
    text = f"  |{progress_bar}| {percent}% ({file_count}/{total_count})\r"
    sys.stdout.write(text)



import sys
import logging

from logging import CRITICAL, ERROR, DEBUG, INFO, WARN, NOTSET

_levels = {
    CRITICAL: "critical",
    ERROR: "error",
    WARN: "warning",
    INFO: "info",
    DEBUG: "debug",
    NOTSET: "none",
}

levels = [name for _, name in _levels.items()]


def setup_logging():
    for level, name in _levels.items():
        logging.addLevelName(level, name)

    handlers = [logging.StreamHandler(sys.stdout)]
    logging.basicConfig(
        handlers=handlers,
        format=("{asctime:^} - {levelname: ^7} - {funcName: ^16} - {message}"),
        style="{",
        datefmt="%m/%d/%Y %H:%M:%S",
    )



base_url = 'https://piccoma.com'
login_url = base_url + '/web/acc/email/signin'
history_url = base_url + '/web/bookshelf/history'
bookmark_url = base_url + '/web/bookshelf/bookmark'
purchase_url = base_url + '/web/bookshelf/purchase'



base_url = 'https://piccoma.com/fr'
login_url = base_url + '/api/auth/signin'
api_url = base_url + '/_next/data/%s/fr'
history_url = '%s/bookshelf/history.json'
bookmark_url = '%s/bookshelf/bookmark.json'
purchase_url = '%s/bookshelf/purchase.json'


#!/usr/bin/env python

import re
import logging

from typing import Mapping, Union, Dict

from pyccoma import Scraper
from pyccoma.exceptions import PyccomaError, PageError, LoginError
from pyccoma.helpers import trunc_title

from pyccoma.jp.urls import (
    base_url,
    login_url,
    history_url,
    bookmark_url,
    purchase_url,
)

log = logging.getLogger(__name__)


class Pyccoma(Scraper):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._etype = {
            "manga": "V",
            "smartoon": "E",
            "novel": "E"
        }

    @property
    def manga(self) -> str:
        return self._etype['manga']

    @property
    def smartoon(self) -> str:
        return self._etype['smartoon']

    @property
    def novel(self) -> str:
        return self._etype['novel']
    
    @manga.setter
    def manga(self, value: str) -> None:
        if value in ('volume', 'episode'):
            self._etype['manga'] = value.upper()[0]
        else:
            raise ValueError("Invalid type.")

    @smartoon.setter
    def smartoon(self, value: str) -> None:
        if value in ('volume', 'episode'):
            self._etype['smartoon'] = value.upper()[0]
        else:
            raise ValueError("Invalid type.")

    @novel.setter
    def novel(self, value: str) -> None:
        if value in ('volume', 'episode'):
            self._etype['novel'] = value.upper()[0]
        else:
            raise ValueError("Invalid type.") 

    def get_login_status(self) -> bool:
        is_login = self.parse_page(login_url).xpath(
            '//script[contains(text(), "login")]/text()'
        )[0].split(":")[1].split(",")[0].strip().title()
        return eval(is_login)

    def login(self, email: str, password: str) -> None:
        try:
            session = self.session.get(login_url, headers=self.headers)
            csrf = self.parse(session.text).xpath(
                f'//input[@name = "csrfmiddlewaretoken"]/@value'
            )

            params = {
                'csrfmiddlewaretoken': csrf,
                'next_url': '/web/',
                'email': email,
                'password': password
            }
            self.session.post(
                login_url,
                data=params,
                cookies=session.cookies,
                headers=self.headers
            )

            if self.get_login_status():
                self._is_login = True
                log.info(f"Successfully logged in as {email}")
            else:
                self._is_login = False
                log.error(f"Failed to log in as {email}")

        except Exception:
            raise SystemExit("Failed to establish connection to server.")

    def get_list(self, url: str) -> Mapping[int, Dict[str, bool]]:
        try:
            if url.endswith('V'):
                return self.get_volume_list(url)
            elif url.endswith('E'):
                return self.get_episode_list(url)
            else:
                raise ValueError("Invalid input.")

        except Exception:
            raise PageError(url)

    def get_episode_list(
        self,
        url: str
    ) -> Mapping[int, Dict[str, Union[str, bool]]]:
        try:
            page = self.parse_page(url).xpath('//ul[@id="js_episodeList"]')[0]
            episode_title = [
                title.text_content() for title in page.xpath(
                    './/div[@class="PCM-epList_title"]/h2'
                )
            ]
            episode_id = [id for id in page.xpath('./li/a/@data-episode_id')]
            series_id = url.split('/')[-2]

            episode_link = [
                "https://piccoma.com/web/viewer/{0}/{1}"
                .format(series_id, id) for id in episode_id
            ]

            status = [_status for _status in page.xpath('./li')]

            episodes = {
                id: {
                    'title': title,
                    'url': link,
                    'is_free': True if _status.find_class('PCM-epList_status_free') else False,  # noqa:E501
                    'is_zero_plus': True if _status.find_class('PCM-epList_status_zeroPlus') else False,  # noqa:E501
                    'is_read_for_free': True if _status.find_class('PCM-epList_status_waitfreeRead') else False,  # noqa:E501
                    'is_already_read': True if _status.find_class('PCM-epList_read') else False,  # noqa:E501
                    'is_wait_for_free': True if _status.find_class('PCM-epList_status_webwaitfree') else False,  # noqa:E501
                    'is_purchased': True if _status.find_class('PCM-epList_status_buy') else False  # noqa:E501
                }
                for id, title, link, _status in zip(
                    episode_id,
                    episode_title,
                    episode_link,
                    status
                )
            }

            if not episodes:
                log.debug(f"No episodes found on {url}")

            return episodes

        except KeyError:
            log.error("Unable to fetch episode list.")
        except AttributeError:
            raise PageError(url)
        except Exception as err:
            raise PyccomaError(err)

    def get_volume_list(
        self,
        url: str
    ) -> Mapping[int, Dict[str, Union[str, bool]]]:
        try:
            page = self.parse_page(url).xpath('//ul[@id="js_volumeList"]')[0]
            volume_title = [
                title.text_content() for title in page.xpath(
                    './/div[@class="PCM-prdVol_title"]//h2'
                )
            ]
            series_id = url.split('/')[-2]
            volume_id = [
                [id for id in links.xpath('./a/@data-episode_id')]
                for links in page.xpath('//div[@class="PCM-prdVol_btns"]')
            ]
            volume_link = [
                "https://piccoma.com/web/viewer/{0}/{1}"
                .format(series_id, id[0]) for id in volume_id
            ]

            xpath_status = './li'
            status = [_status for _status in page.xpath(f'{xpath_status}')]

            volumes = {
                id + 1: {
                    'title': title,
                    'url': link,
                    'is_free': True if _status.find_class('PCM-prdVol_freeBtn') else False,  # noqa:E501
                    'is_read_for_free': True if (_status.find_class('PCM-prdVol_readBtn') and _status.find_class('PCM-prdVol_campaign_free')) else False,  # noqa:E501
                    'is_already_read': True if _status.find_class('PCM-volList_read') else False,  # noqa:E501
                    'is_wait_for_free': True if _status.find_class('PCM-prdVol_campaign_free') else False,  # noqa:E501
                    'is_purchased': True if _status.find_class('PCM-prdVol_readBtn') else False  # noqa:E501
                }
                for id, (title, link, _status) in enumerate(
                    zip(
                        volume_title,
                        volume_link,
                        status
                    )
                )
            }

            if not volumes:
                log.debug(f"No volumes found on {url}")

            return volumes

        except ValueError:
            log.error("Invalid url, unable to fetch volume list.")
        except AttributeError:
            raise PageError(url)
        except Exception as err:
            raise PyccomaError(err)

    def get_bdata(self, url: str) -> Dict[str, str]:
        try:
            page = self.parse_page(url).xpath(
                '//section[@class="PCM-productTile"]'
            )[0]
            product_title = [
                title.text for title in page.xpath('.//p//span') if title.text
            ]
            product_type = [
                self.smartoon if 'PCM-stt_smartoon' in _type else
                self.novel if 'PCM-stt_novel' in _type else self.manga
                for _type in [
                    type.classes for type in page.xpath('.//li')
                ]
            ]
            product_link = [
                base_url + url + "/episodes?etype="
                for url in page.xpath('//a[@class="PCM-product"]/@href')
            ]
            items = {
                title: link + type for title, link, type in zip(
                    product_title, product_link, product_type
                )
            }
            return items

        except Exception:
            log.error("Failed to parse library.")

    def get_pdata(self, url: str) -> Dict[str, Union[str, bool]]:
        try:
            page = self.parse_page(url)

            log.info(f"Parsing data from {url}")

            title = page.xpath('//title')[0].text_content().split("｜")[1]

            page = page.xpath('//script[contains(text(), "pdata")]/text()')[0]
            ep_title = page.split("'title'")[1].split("'")[1].strip()

            pattern = r"(?<=:')[^']+(?=')"
            images = ["https:" + image for image in re.findall(pattern, page)]
            pdata = {
                'title': title if not self.omit_author else trunc_title(title),
                'ep_title': ep_title,
                'img': images
            }

            return pdata

        except TypeError or IndexError:
            log.error("Unable to fetch page data.")

    def get_checksum(self, img_url: str) -> str:
        return img_url.split('/')[-2]

    def get_history(self) -> Dict[str, str]:
        if self._is_login:
            return self.get_bdata(history_url)
        else:
            raise LoginError

    def get_bookmark(self) -> Dict[str, str]:
        if self._is_login:
            return self.get_bdata(bookmark_url)
        else:
            raise LoginError

    def get_purchase(self) -> Dict[str, str]:
        if self._is_login:
            return self.get_bdata(purchase_url)
        else:
            raise LoginError



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

requirements = [
    "lxml>=4.7.1",
    "requests==2.26.0",
    "image-scramble==2.0.1"
]

setup(
    name="pyccoma",
    version="0.5.1",
    author="catsital",
    author_email="catshital@gmail.com",
    description="Scrape and download from Piccoma Japan and France.",
    entry_points={"console_scripts": ["pyccoma=pyccoma.__main__:main"],},
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/catsital/pyccoma",
    project_urls={
        "Bug Tracker": "https://github.com/catsital/pyccoma/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    zip_safe=False
)



class PyccomaError(Exception):
    """Any error caused by Pyccoma will be caught
       with this exception."""


class PageError(PyccomaError):
    def __init__(self, url):
        err = f"Unable to access page on {url}"
        Exception.__init__(self, err)


class LoginError(PyccomaError):
    def __init__(self):
        err = "Login required."
        Exception.__init__(self, err)
